"""sql compatibility

Revision ID: fc07e3fa0086
Revises: 7f317474332d
Create Date: 2017-01-30 16:09:53.367166

"""

# revision identifiers, used by Alembic.
revision = "fc07e3fa0086"
down_revision = "7f317474332d"

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("session", schema=None) as batch_op:
        batch_op.alter_column(
            "ip",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=256),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "ua",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=2048),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "user",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=256),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "uuid",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=256),
            existing_nullable=True,
        )

    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.alter_column(
            "task",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=256),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "user",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=256),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "uuid",
            existing_type=sa.VARCHAR(),
            type_=sa.String(length=256),
            existing_nullable=True,
        )

    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.alter_column(
            "uuid",
            existing_type=sa.String(length=256),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "user",
            existing_type=sa.String(length=256),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "task",
            existing_type=sa.String(length=256),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )

    with op.batch_alter_table("session", schema=None) as batch_op:
        batch_op.alter_column(
            "uuid",
            existing_type=sa.String(length=256),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "user",
            existing_type=sa.String(length=256),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "ua",
            existing_type=sa.String(length=2048),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )
        batch_op.alter_column(
            "ip",
            existing_type=sa.String(length=256),
            type_=sa.VARCHAR(),
            existing_nullable=True,
        )

    ### end Alembic commands ###



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.backend.parallel
    :platform: Unix
    :synopsis: Burp-UI parallel backend module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import re
import json
import ssl
import time
import trio
import struct

from asyncio import iscoroutinefunction
from functools import partial

from .burp2 import Burp as Burp2
from .interface import BUIbackend, BUIBACKEND_INTERFACE_METHODS
from .utils.constant import BURP_STATUS_FORMAT_V2
from ..parser.burp2 import Parser
from ...exceptions import BUIserverException
from ...decorators import implement, usetriorun
from ...utils import utc_to_local
from ..._compat import to_unicode, to_bytes
from ...tools.logging import logger

BUI_DEFAULTS = {
    "Parallel": {
        "host": "::1",
        "port": 11111,
        "ssl": True,
        "password": "password123456",
        "timeout": 15,
        "concurrency": 2,
        "init_wait": 15,
    },
}


class Connector:
    logger = logger

    def __init__(self, conf):
        """Connector client

        :param conf: Configuration to use
        :type conf: :class:`burpui.config.BUIConfig`
        """
        self.host = conf.safe_get("host", section="Parallel", defaults=BUI_DEFAULTS)
        self.port = conf.safe_get(
            "port", "integer", section="Parallel", defaults=BUI_DEFAULTS
        )
        self.ssl = conf.safe_get(
            "ssl", "boolean", section="Parallel", defaults=BUI_DEFAULTS
        )
        self.password = conf.safe_get(
            "password", section="Parallel", defaults=BUI_DEFAULTS
        )
        self.timeout = conf.safe_get(
            "timeout", "integer", section="Parallel", defaults=BUI_DEFAULTS
        )

        self.logger.debug(f"Monitor {self.host}:{self.port} - ssl: {self.ssl}")

        self.connected = False

    async def conn(self):
        try:
            if self.ssl:
                ctx = ssl.SSLContext()
                ctx.verify_mode = ssl.CERT_NONE
                ctx.check_hostname = False
                ctx.options |= (
                    ssl.OP_NO_SSLv2
                    | ssl.OP_NO_SSLv3
                    | ssl.OP_NO_TLSv1
                    | ssl.OP_NO_TLSv1_1
                )  # RFC 7540 Section 9.2: MUST be TLS >=1.2
                ctx.options |= (
                    ssl.OP_NO_COMPRESSION
                )  # RFC 7540 Section 9.2.1: MUST disable compression
                ctx.load_default_certs()
                self.client_stream = await trio.open_ssl_over_tcp_stream(
                    self.host, self.port, ssl_context=ctx
                )
            else:
                self.client_stream = await trio.open_tcp_stream(self.host, self.port)
        except OSError as exc:
            raise BUIserverException(str(exc))

        self.logger.debug("Connected")
        self.connected = True
        return self.client_stream

    async def _send(self, data):
        data = to_bytes(data)
        length = struct.pack("!Q", len(data))
        await self.client_stream.send_all(length)
        self.logger.debug(f"Sending: {data!r}")
        await self.client_stream.send_all(data)

    async def _do_process(self, data):
        res = "[]"
        await self._send(data)
        tmp = await self.client_stream.receive_some(2)
        tmp = to_unicode(tmp)
        if tmp == "ER":
            lengthbuf = await self.client_stream.receive_some(8)
            (length,) = struct.unpack("!Q", lengthbuf)
            err = await self.receive_all(length)
            err = to_unicode(err)
            raise BUIserverException(err)
        if tmp != "OK":
            self.logger.debug("Ooops, unsuccessful!")
            return res
        self.logger.debug("Data sent successfully")
        lengthbuf = await self.client_stream.receive_some(8)
        (length,) = struct.unpack("!Q", lengthbuf)
        res = await self.receive_all(length)
        res = to_unicode(res)
        self.logger.debug(f"Received: {res!r}")
        return res

    async def _process(self, data):
        if not self.connected:
            await self.conn()
            async with self.client_stream:
                result = await self._do_process(data)
            self.connected = False
        else:
            result = await self._do_process(data)
        return result

    async def status(self, query, timeout=None, cache=True):

        request = {
            "query": query,
            "timeout": timeout,
            "cache": cache,
            "password": self.password,
        }
        request = json.dumps(request)
        result = await self._process(request)
        return json.loads(result)

    async def request(self, func, *args, **kwargs):
        req = {
            "func": func,
            "args": args,
            "kwargs": kwargs,
            "password": self.password,
        }
        req = json.dumps(req)
        result = await self._process(req)
        return result

    async def receive_all(self, length=1024, bsize=None):
        buf = b""
        bsize = bsize if bsize is not None else 1024
        bsize = min(bsize, length)
        received = 0
        tries = 0
        while received < length:
            newbuf = await self.client_stream.receive_some(bsize)
            if not newbuf:
                # 3 successive read failure => raise exception
                if tries > 3:
                    raise IOError("Unable to read full response")
                tries += 1
                trio.sleep(0.1)
                continue
            # reset counter
            tries = 0
            buf += newbuf
            received += len(newbuf)
        return buf


# Asynchronous functions for the bui-monitor pool
class AsyncBurpMixin:
    async def _async_statistics(self, agent=None):
        return json.loads(await self._async_request("statistics"))

    async def _async_get_client_version(self, agent=None):
        if self._client_version is None:
            try:
                self._client_version = await self._async_request("client_version")
            except BUIserverException:
                return ""
        return self._client_version or ""

    async def _async_get_server_version(self, agent=None):
        if self._server_version is None:
            try:
                self._server_version = await self._async_request("server_version")
            except BUIserverException:
                return ""
        return self._server_version or ""

    async def _async_status(self, query="c:\n", timeout=None, cache=True, agent=None):
        try:
            connector = Connector(self.conf)
            return await connector.status(query, timeout, cache)
        except (OSError, IOError) as exc:
            raise BUIserverException(str(exc))
        if not self._ready:
            self.init_all()

    async def _async_request(self, func, *args, **kwargs):
        try:
            connector = Connector(self.conf)
            return await connector.request(func, *args, **kwargs)
        except (OSError, IOError) as exc:
            raise BUIserverException(str(exc))

    async def _async_get_backup_logs(
        self, number, client, forward=False, deep=False, store=None, limit=None
    ):
        async def _do_stuff():
            nonlocal client
            nonlocal number
            nonlocal forward
            nonlocal deep
            bucket1 = []
            bucket2 = []
            ret = {}
            query = await self._async_status("c:{0}:b:{1}\n".format(client, number))
            if not query:
                return ret
            try:
                logs = query["clients"][0]["backups"][0]["logs"]["list"]
            except KeyError:
                self.logger.warning("No logs found")
                return ret
            async with trio.open_nursery() as nursery:
                if "backup_stats" in logs:
                    nursery.start_soon(
                        self._async_parse_backup_stats,
                        number,
                        client,
                        forward,
                        None,
                        bucket1,
                    )
                if "backup" in logs and deep:
                    nursery.start_soon(
                        self._async_parse_backup_log, number, client, bucket2
                    )

            if bucket1:
                ret = bucket1[0]
            if bucket2:
                ret.update(bucket2[0])

            ret["encrypted"] = False
            if "files_enc" in ret and ret["files_enc"]["total"] > 0:
                ret["encrypted"] = True
            return ret

        if limit is not None:
            async with limit:
                res = await _do_stuff()
        else:
            res = await _do_stuff()

        if store is not None:
            store.append(res)
        else:
            return res

    async def _async_get_all_backup_logs(self, client, forward=False, deep=False):
        ret = []
        backups = await self._async_get_client(client)
        queue = []
        limit = trio.CapacityLimiter(self.concurrency)
        async with trio.open_nursery() as nursery:
            for back in backups:
                nursery.start_soon(
                    self._async_get_backup_logs,
                    back["number"],
                    client,
                    forward,
                    deep,
                    queue,
                    limit,
                )

        ret = sorted(queue, key=lambda x: x["number"])
        return ret

    async def _async_parse_backup_log(self, number, client, bucket=None):
        query = await self._async_status(
            "c:{0}:b:{1}:l:backup\n".format(client, number)
        )
        res = self._do_parse_backup_log(query, client)
        if bucket is not None:
            bucket.append(res)
        return res

    async def _async_parse_backup_stats(
        self, number, client, forward=False, agent=None, bucket=None
    ):
        """The :func:`burpui.misc.backend.parallel.Burp._async_parse_backup_stats`
        function is used to parse the burp logs.

        :param number: Backup number to work on
        :type number: int

        :param client: Client name to work on
        :type client: str

        :param forward: Is the client name needed in later process
        :type forward: bool

        :param agent: What server to ask (only in multi-agent mode)
        :type agent: str

        :returns: Dict containing the backup log
        """
        backup = {"os": await self._async_guess_os(client), "number": int(number)}
        if forward:
            backup["name"] = client
        query = await self._async_status(
            "c:{0}:b:{1}:l:backup_stats\n".format(client, number)
        )
        ret = self._do_parse_backup_stats(query, backup, number, client, forward, agent)
        if bucket is not None:
            bucket.append(ret)
        return ret

    async def _async_get_clients_report(self, clients, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_clients_report`"""

        async def __compute_client_report(cli, queue, limit):
            async with limit:
                if not cli:
                    return
                client = await self._async_get_client(cli["name"])
                if not client or not client[-1]:
                    return
                stats = await self._async_get_backup_logs(
                    client[-1]["number"], cli["name"]
                )
                queue.append((cli, client, stats))

        data = []
        limiter = trio.CapacityLimiter(self.concurrency)

        async with trio.open_nursery() as nursery:
            for client in clients:
                nursery.start_soon(__compute_client_report, client, data, limiter)

        return self._do_get_clients_report(data)

    async def _async_get_counters(self, name=None, agent=None):
        ret = {}
        query = await self._async_status("c:{0}\n".format(name), cache=False)
        # check the status returned something
        if not query:
            return ret

        try:
            client = query["clients"][0]
        except KeyError:
            self.logger.warning("Client not found")
            return ret
        return self._do_get_counters(client)

    async def _async_is_backup_running(self, name=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_backup_running`
        """
        if not name:
            return False
        try:
            query = await self._async_status("c:{0}\n".format(name))
        except BUIserverException:
            return False
        return self._do_is_backup_running(query)

    async def _async_is_one_backup_running(self, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_one_backup_running`
        """
        ret = []
        try:
            clients = await self._async_get_all_clients(deep=False, last_attempt=False)
        except BUIserverException:
            return ret
        return self._do_is_one_backup_running(clients)

    async def _async_get_last_backup(self, name, working=True):
        """Return the last backup of a given client

        :param name: Name of the client
        :type name: str

        :param working: Also return uncomplete backups
        :type working: bool

        :returns: The last backup
        """
        try:
            clients = await self._async_status("c:{}".format(name))
            client = clients["clients"][0]
            i = 0
            while True:
                ret = client["backups"][i]
                if not working and "working" in ret["flags"]:
                    i += 1
                    continue
                return ret
        except (KeyError, IndexError, BUIserverException):
            return None

    async def _async_guess_os(self, name):
        """Return the OS of the given client based on the magic *os* label

        :param name: Name of the client
        :type name: str

        :returns: The guessed OS of the client

        ::

            grep label /etc/burp/clientconfdir/toto
            label = os: Darwin OS
        """
        ret = "Unknown"
        if name in self._os_cache:
            return self._os_cache[name]

        labels = await self._async_get_client_labels(name)
        OSES = []

        for label in labels:
            if re.match("os:", label, re.IGNORECASE):
                _os = label.split(":", 1)[1].strip()
                if _os not in OSES:
                    OSES.append(_os)

        if OSES:
            ret = OSES[-1]
        else:
            # more aggressive check
            last = await self._async_get_last_backup(name, False)
            if last:
                try:
                    tree = await self._async_get_tree(name, last["number"])

                    if tree[0]["name"] != "/":
                        ret = "Windows"
                    else:
                        ret = "Unix/Linux"
                except (IndexError, KeyError, BUIserverException):
                    pass

        self._os_cache[name] = ret
        return ret

    async def _async_get_all_clients(self, agent=None, deep=True, last_attempt=True):
        ret = []
        query = await self._async_status()
        if not query or "clients" not in query:
            return ret

        async def __compute_client_data(client, queue, limit):
            async with limit:
                cli = {}
                cli["name"] = client["name"]
                cli["state"] = self._status_human_readable(client["run_status"])
                infos = client["backups"]
                if cli["state"] in ["running"]:
                    cli["last"] = "now"
                    cli["last_attempt"] = "now"
                elif not infos:
                    cli["last"] = "never"
                    cli["last_attempt"] = "never"
                else:
                    convert = True
                    infos = infos[0]
                    server_version = await self._async_get_server_version()
                    if server_version and server_version < BURP_STATUS_FORMAT_V2:
                        cli["last"] = infos["timestamp"]
                        convert = False
                    # only do deep inspection when server >= BURP_STATUS_FORMAT_V2
                    if deep:
                        logs = await self._async_get_backup_logs(
                            infos["number"], client["name"]
                        )
                        cli["last"] = logs["start"]
                    else:
                        cli["last"] = utc_to_local(infos["timestamp"])
                    if last_attempt:
                        last_backup = await self._async_get_last_backup(client["name"])
                        if convert:
                            cli["last_attempt"] = utc_to_local(last_backup["timestamp"])
                        else:
                            cli["last_attempt"] = last_backup["timestamp"]
                queue.append(cli)

        clients = query["clients"]
        limiter = trio.CapacityLimiter(self.concurrency)

        async with trio.open_nursery() as nursery:
            for client in clients:
                nursery.start_soon(__compute_client_data, client, ret, limiter)

        return ret

    async def _async_get_client_status(self, name=None, agent=None):
        ret = {}
        if not name:
            return ret
        query = await self._async_status("c:{0}\n".format(name))
        if not query:
            return ret
        try:
            client = query["clients"][0]
        except (KeyError, IndexError):
            self.logger.warning("Client not found")
            return ret
        return self._do_get_client_status(client)

    async def _async_get_client_filtered(
        self, name=None, limit=-1, page=None, start=None, end=None, agent=None
    ):
        ret = []
        if not name:
            return ret
        query = await self._async_status("c:{0}\n".format(name))
        if not query:
            return ret
        try:
            backups = query["clients"][0]["backups"]
        except (KeyError, IndexError):
            self.logger.warning("Client not found")
            return ret

        async def __parse_log(backup, client, back, ret, limiter):
            async with limiter:
                append = True
                log = await self._async_get_backup_logs(backup["number"], client)
                try:
                    back["encrypted"] = log["encrypted"]
                    try:
                        back["received"] = log["received"]
                    except KeyError:
                        back["received"] = 0
                    try:
                        back["size"] = log["totsize"]
                    except KeyError:
                        back["size"] = 0
                    back["end"] = log["end"]
                    # override date since the timestamp is odd
                    back["date"] = log["start"]
                except Exception:
                    self.logger.warning("Unable to parse logs")
                    append = False

                if append:
                    ret.append(back)

        queue = []
        limiter = trio.CapacityLimiter(self.concurrency)

        async with trio.open_nursery() as nursery:
            for idx, backup in enumerate(backups):
                back = {}
                # skip the first elements if we are in a page
                if page and page > 1 and limit > 0:
                    if idx < (page - 1) * limit:
                        continue

                # skip running backups since data will be inconsistent
                if "flags" in backup and "working" in backup["flags"]:
                    continue
                back["number"] = backup["number"]
                if "flags" in backup and "deletable" in backup["flags"]:
                    back["deletable"] = True
                else:
                    back["deletable"] = False
                back["date"] = backup["timestamp"]
                # skip backups before "start"
                if start and backup["timestamp"] < start:
                    continue
                # skip backups after "end"
                if end and backup["timestamp"] > end:
                    continue

                nursery.start_soon(__parse_log, backup, name, back, queue, limiter)

                # stop after "limit" elements
                if page and page > 1 and limit > 0:
                    if idx >= page * limit:
                        break
                elif limit > 0 and idx >= limit:
                    break

        # Here we need to reverse the array so the backups are sorted by num
        # ASC
        ret = sorted(queue, key=lambda x: x["number"])
        return ret

    async def _async_get_tree(
        self, name=None, backup=None, root=None, level=-1, agent=None
    ):
        ret = []
        if not name or not backup:
            return ret
        if not root:
            top = ""
        else:
            top = to_unicode(root)

        # we know this operation may take a while so we arbitrary increase the
        # read timeout
        timeout = None
        if top == "*":
            timeout = max(self.timeout, 300)

        query = await self._async_status(
            "c:{0}:b:{1}:p:{2}\n".format(name, backup, top), timeout
        )
        return self._format_tree(query, top, level)

    async def _async_get_client(self, name=None, agent=None):
        return await self._async_get_client_filtered(name)

    async def _async_get_client_labels(self, client=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_client_labels`
        """
        ret = []
        if not client:
            return ret
        # micro optimization since the status results are cached in memory for a
        # couple seconds, using the same global query and iterating over it
        # will be more efficient than filtering burp-side
        query = await self._async_status("c:\n")
        if not query:
            return ret
        try:
            for cli in query["clients"]:
                if cli["name"] == client:
                    return cli["labels"]
        except KeyError:
            return ret

    async def _async_restore_files(
        self,
        name=None,
        backup=None,
        files=None,
        strip=None,
        archive="zip",
        password=None,
        agent=None,
    ):
        return await trio.to_thread.run_sync(
            Burp2.restore_files,
            self,
            name,
            backup,
            files,
            strip,
            archive,
            password,
            agent,
        )

    async def _async_is_backup_deletable(self, name=None, backup=None, agent=None):
        if not name or not backup:
            return False
        query = await self._async_status("c:{0}:b:{1}\n".format(name, backup))
        if not query:
            return False
        return self._do_is_backup_deletable(query)


# Some functions are the same as in Burp1 backend
class Burp(Burp2, AsyncBurpMixin):
    """The :class:`burpui.misc.backend.parallel.Burp` class provides a consistent
    backend for ``burp-2`` servers through the bui-monitor pool. It is also able to
    perform some operations asynchronously to speedup the whole API.

    It extends the :class:`burpui.misc.backend.burp2.Burp` class because a few
    functions can be reused. The rest is just overridden.

    :param server: ``Burp-UI`` server instance in order to access logger
                   and/or some global settings
    :type server: :class:`burpui.engines.server.BUIServer`

    :param conf: Configuration to use
    :type conf: :class:`burpui.config.BUIConfig`
    """

    # backend version
    _vers = 2
    # cache to store the guessed OS
    _os_cache = {}

    _client_version = None
    _server_version = None
    _batch_list_supported = None

    _ready = False
    parser = None

    def __init__(self, server=None, conf=None):
        """
        :param server: ``Burp-UI`` server instance in order to access logger
                       and/or some global settings
        :type server: :class:`burpui.engines.server.BUIServer`

        :param conf: Configuration to use
        :type conf: :class:`burpui.config.BUIConfig`
        """

        BUIbackend.__init__(self, server, conf)

        self.conf = conf
        self.concurrency = conf.safe_get(
            "concurrency", "integer", "Parallel", BUI_DEFAULTS
        )
        self.init_wait = conf.safe_get("init_wait", "integer", "Parallel", BUI_DEFAULTS)

        if os.getenv("BUI_MODE", "") == "celery":
            # we cap the concurrency level in order not to prevent our main server
            # to talk to the monitor
            self.concurrency = max(1, self.concurrency // 2)

        self.logger.info("burp conf cli: {}".format(self.burpconfcli))
        self.logger.info("burp conf srv: {}".format(self.burpconfsrv))
        self.logger.info("command timeout: {}".format(self.timeout))
        self.logger.info("tmpdir: {}".format(self.tmpdir))
        self.logger.info("zip64: {}".format(self.zip64))
        self.logger.info("includes: {}".format(self.includes))
        self.logger.info("enforce: {}".format(self.enforce))
        self.logger.info("revoke: {}".format(self.revoke))
        self.logger.info("concurrency: {}".format(self.concurrency))
        self.logger.info("init_wait: {}".format(self.init_wait))

        if self.init_wait:
            exc = None
            for i in range(self.init_wait):
                connector = Connector(conf)
                try:
                    self.logger.warning(
                        "monitor not ready, waiting for it... {}/{}".format(
                            i, self.init_wait
                        )
                    )
                    trio.run(connector.conn)
                    if connector.connected:
                        trio.run(connector._send, "RE")
                        break
                except BUIserverException as eee:
                    exc = eee
                time.sleep(1)
            else:
                self.logger.error(f"monitor not ready, giving up!: {exc}")

        try:
            stats = self.statistics()
            if "alive" in stats and stats["alive"]:
                self.init_all()
        except BUIserverException:
            pass

    def init_all(self):
        self._ready = True
        self.parser = Parser(self)

    @property
    def client_version(self):
        return self.get_client_version()

    @property
    def server_version(self):
        return self.get_server_version()

    @property
    def batch_list_supported(self):
        if self._batch_list_supported is None:
            self._batch_list_supported = json.loads(
                trio.run(self._async_request, "batch_list_supported")
            )
        return self._batch_list_supported

    @usetriorun
    def statistics(self, agent=None):
        return trio.run(self._async_statistics)

    @usetriorun
    def get_client_version(self, agent=None):
        if self._client_version is None:
            self._client_version = trio.run(self._async_get_client_version)
        return self._client_version or ""

    @usetriorun
    def get_server_version(self, agent=None):
        if self._server_version is None:
            self._server_version = trio.run(self._async_get_server_version)
        return self._server_version or ""

    @usetriorun
    def status(self, query="c:\n", timeout=None, cache=True, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.status`"""
        return trio.run(self._async_status, query, timeout, cache)

    def get_backup_logs(self, number, client, forward=False, deep=False, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_backup_logs`
        """
        if not client or not number:
            return {} if number and number != -1 else []

        if number == -1:
            return trio.run(self._async_get_all_backup_logs, client, forward, deep)
        return trio.run(self._async_get_backup_logs, number, client, forward, deep)

    def _parse_backup_log(self, number, client):
        """The :func:`burpui.misc.backend.burp2.Burp._parse_backup_log`
        function helps you determine if the backup is protocol 2 or 1 and various
        useful details.

        :param number: Backup number to work on
        :type number: int

        :param client: Client name to work on
        :type client: str

        :returns: a dict with some useful details
        """
        return trio.run(self._async_parse_backup_log, number, client)

    @usetriorun
    def get_clients_report(self, clients, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_clients_report`"""
        return trio.run(self._async_get_clients_report, clients)

    @usetriorun
    def get_counters(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_counters`"""
        return trio.run(self._async_get_counters, name)

    @usetriorun
    def is_backup_running(self, name=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_backup_running`
        """
        return trio.run(self._async_is_backup_running, name)

    @usetriorun
    def is_one_backup_running(self, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_one_backup_running`
        """
        return trio.run(self._async_is_one_backup_running)

    @usetriorun
    def _get_last_backup(self, name, working=True):
        """Return the last backup of a given client

        :param name: Name of the client
        :type name: str

        :param working: Also return uncomplete backups
        :type working: bool

        :returns: The last backup
        """
        return trio.run(self._async_get_last_backup, name, working)

    @usetriorun
    def _guess_os(self, name):
        """Return the OS of the given client based on the magic *os* label

        :param name: Name of the client
        :type name: str

        :returns: The guessed OS of the client

        ::

            grep label /etc/burp/clientconfdir/toto
            label = os: Darwin OS
        """
        return trio.run(self._async_guess_os, name)

    def get_all_clients(self, agent=None, last_attempt=True):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_all_clients`
        """
        # don't need async processing if burp-server < BURP_STATUS_FORMAT_V2
        if not self.deep_inspection or (
            self.server_version and self.server_version < BURP_STATUS_FORMAT_V2
        ):
            return Burp2.get_all_clients(self, last_attempt=last_attempt)
        # the deep inspection can take advantage of async processing
        callback = partial(self._async_get_all_clients, last_attempt=last_attempt)
        return trio.run(callback)

    @usetriorun
    def get_client_status(self, name=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_client_status`
        """
        return trio.run(self._async_get_client_status, name)

    @usetriorun
    def get_client(self, name=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client`"""
        return trio.run(self._async_get_client, name)

    @usetriorun
    def get_client_filtered(
        self, name=None, limit=-1, page=None, start=None, end=None, agent=None
    ):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_client_filtered`"""
        return trio.run(self._async_get_client_filtered, name, limit, page, start, end)

    @usetriorun
    def is_backup_deletable(self, name=None, backup=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.is_backup_deletable`
        """
        return trio.run(self._async_is_backup_deletable, name, backup)

    @usetriorun
    def get_tree(self, name=None, backup=None, root=None, level=-1, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_tree`"""
        return trio.run(self._async_get_tree, name, backup, root, level)

    @usetriorun
    def get_client_labels(self, client=None, agent=None):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_client_labels`
        """
        return trio.run(self._async_get_client_labels, client)

    @usetriorun
    def restore_files(
        self,
        name=None,
        backup=None,
        files=None,
        strip=None,
        archive="zip",
        password=None,
        agent=None,
    ):
        return trio.run(
            self._async_restore_files,
            name,
            backup,
            files,
            strip,
            archive,
            password,
            agent,
        )

    # Same as in Burp1 backend

    # def read_conf_cli(self, agent=None):

    # def read_conf_srv(self, agent=None):

    # def store_conf_cli(self, data, agent=None):

    # def store_conf_srv(self, data, agent=None):

    # def get_parser_attr(self, attr=None, agent=None):


# Make every "Burp" method async
class AsyncBurp(Burp):
    @property
    async def batch_list_supported(self):
        if self._batch_list_supported is None:
            self._batch_list_supported = json.loads(
                await self._async_request("batch_list_supported")
            )
        return self._batch_list_supported

    # this method must not be async!
    @implement
    def statistics(self, agent=None):
        return Burp.statistics(self)

    # this method must not be async!
    @implement
    def get_parser(self, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_parser`"""
        return Burp.get_parser(self)

    @property
    def client_version(self):
        return trio.run(self.get_client_version)

    @property
    def server_version(self):
        return trio.run(self.get_server_version)

    @implement
    async def get_client_version(self, agent=None):
        if self._client_version is None:
            self._client_version = await self._async_request("client_version")
        return self._client_version

    @implement
    async def get_server_version(self, agent=None):
        if self._server_version is None:
            self._server_version = await self._async_request("server_version")
        return self._server_version

    @implement
    async def get_backup_logs(
        self, number, client, forward=False, deep=False, agent=None
    ):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_backup_logs`
        """
        if not client or not number:
            return {} if number and number != -1 else []

        if number == -1:
            return await self._async_get_all_backup_logs(client, forward, deep)
        return await self._async_get_backup_logs(number, client, forward, deep)

    @implement
    async def get_all_clients(self, agent=None, last_attempt=True):
        """See
        :func:`burpui.misc.backend.interface.BUIbackend.get_all_clients`
        """
        return await self._async_get_all_clients(last_attempt=last_attempt)

    @implement
    async def get_attr(self, name, default=None, agent=None):
        """See :func:`burpui.misc.backend.interface.BUIbackend.get_attr`"""
        try:
            try:
                return await getattr(self, name, default)
            except TypeError:
                return getattr(self, name, default)
        except AttributeError:
            return default

    def __getattribute__(self, name):
        if name in BUIBACKEND_INTERFACE_METHODS or name in ["_guess_os"]:
            wrap = False
            proxy = True
            func = None
            try:
                func = object.__getattribute__(self, name)
                proxy = not getattr(func, "__ismethodimplemented__", False)
                wrap = getattr(func, "__isusingtriorun__", False)
            except:
                pass
            if func and not callable(func):
                self.logger.debug(f"{func} is not a function")
                return func
            self.logger.debug(f"async func: {func} - proxy: {proxy}, wrap: {wrap}")
            if wrap:
                realname = (
                    f"_async_{name}" if not name.startswith("_") else f"_async{name}"
                )
                realfunc = object.__getattribute__(self, realname)
                return ProxyAsyncCall(realfunc, realname, self)
            if proxy:
                return ProxyAsyncCall(func, name, self)
            elif func:
                return func
        return object.__getattribute__(self, name)


class ProxyAsyncCall(object):
    """Class to dispatch call of unknown methods in order to dynamically
    implements their async variant."""

    def __init__(self, func, name, proxy):
        """
        :param proxy: function to proxify
        :type proxy: function

        :param name: Name of the method to proxify
        :type name: str

        :param proxy: Object to proxify
        :type proxy: :class:`Burp`
        """
        self.func = func
        self.name = name
        self.proxy = proxy

    async def __call__(self, *args, **kwargs):
        """This is where the proxy call (and the magic) occurs"""
        # retrieve the original function prototype
        proto = getattr(BUIbackend, self.name, None) or getattr(self.proxy, self.name)
        args_name = list(proto.__code__.co_varnames)
        # skip self
        args_name.pop(0)
        # we transform unnamed arguments to named ones
        # example:
        #     def my_function(toto, tata=None, titi=None):
        #
        #     x = my_function('blah', titi='blih')
        #
        # => {'toto': 'blah', 'titi': 'blih'}
        encoded_args = {}
        for idx, opt in enumerate(args):
            encoded_args[args_name[idx]] = opt
        encoded_args.update(kwargs)

        if iscoroutinefunction(self.func):
            return await self.func(**encoded_args)
        return self.func(**encoded_args)



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.utils
    :platform: Unix
    :synopsis: Burp-UI configuration file parser utilities.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import re
import codecs
import shutil

from copy import copy
from hashlib import md5
from collections import OrderedDict
from glob import glob

from ...utils import NOTIF_ERROR, NOTIF_OK, NOTIF_WARN
from ...security import sanitize_string
from ...datastructures import MultiDict


RESET_IDENTIFIER = "_reset_bui_CUSTOM"
BEGIN_TEMPLATES = "BURP-UI TEMPLATES"
END_TEMPLATES = "END TEMPLATES"


class Option(object):
    """Object representing an option

    :param name: Option name
    :type name: str

    :param value: Option value
    :type value: str
    """

    type = None
    delim = "="
    reset_delim = ":="
    _dirty = False
    idx = -1

    def __init__(self, name, value=None):
        self.name = name
        self.value = value
        self._dirty = True
        # is there a special := syntax for this option
        self._is_reset = []

    @property
    def dirty(self):
        return self._dirty

    def is_reset(self):
        """Check if a special ':=' syntax is required based on the value"""
        if self.idx < 0 or not self._is_reset:
            return False
        if self.idx >= len(self._is_reset):
            self.idx = len(self._is_reset) - 1
        return self._is_reset[self.idx]

    def set_reset(self, value=False):
        """Mark this value as requiring a special ':=' syntax"""
        self._is_reset.append("{}".format(value).lower() == "true")
        self.idx += 1

    def set_resets(self, resets):
        self._is_reset = resets

    def get_resets(self):
        return self._is_reset

    def get_reset(self):
        """Return the reset list/flag"""
        return self.is_reset()

    def update(self, value):
        """Change the option value"""
        self._dirty = True
        self.value = value
        self._is_reset = []

    def clean(self):
        """Mark the option as clean"""
        self._dirty = False

    def parse(self):
        """Parse the option value"""
        return self.value

    def dump(self):
        """Return the option representation to store in configuration file"""
        delim = self.delim
        if self.is_reset():
            delim = self.reset_delim
        return "{} {} {}".format(self.name, delim, self.value)

    def __repr__(self):
        """Option representation"""
        return "{} -> {}".format(self.name, self.parse())

    def __str__(self):
        """Option to string"""
        return self.dump()

    def __eq__(self, other):
        other_name = getattr(other, "name", None)
        other_value = getattr(other, "value", None)
        return self.name == other_name and self.value == other_value


class OptionStr(Option):
    """Option type String

    Example:
        server = toto
    """

    type = "string"


class OptionInt(Option):
    """Option type Integer

    Example:
        port = 1234
    """

    type = "integer"

    def parse(self):
        """Parse the option value"""
        try:
            return int(self.value)
        except (ValueError, TypeError):
            return 0


class OptionBool(Option):
    """Option type Boolean

    Example:
        hardlinked_archive = 1
    """

    type = "boolean"

    def __init__(self, name, value=None):
        self.name = name
        self.value = self._format_value(value)
        self._dirty = True
        self._is_reset = []

    def update(self, value):
        """Change the option value"""
        self._dirty = True
        self.value = self._format_value(value)
        self._is_reset = []

    def _format_value(self, value):
        if self._parse(value):
            return 1
        return 0

    @staticmethod
    def _parse(value):
        try:
            if not value:
                return False
            elif value is True:
                return True
            elif "{}".format(value).lower() == "true":
                return True
            # any string will raise the ValueError
            return int(value) == 1
        except ValueError:
            return False

    def parse(self):
        """Parse the option value"""
        return self._parse(self.value)


class OptionInc(Option):
    """Option type Include

    Example:
        . incexc/windows
    """

    type = "include"
    delim = ""

    def __init__(self, parser, name, value=None, root=None, mode="srv", template=False):
        """
        :param parser: Parser instance
        :type parser: :class:`burpui.misc.parser.burp1.Parser`
        """
        super(OptionInc, self).__init__(name, value)
        self.parser = parser
        self.mode = mode
        self.is_template = template
        self.extended = []
        self._is_reset = []
        self._dirty = True
        self._glob = False
        self._glob_dir = None
        self._glob_mtime = -1
        if root:
            self.root = os.path.dirname(root)
        else:
            self.root = None

    @property
    def dirty(self):
        return self._dirty or self._glob_changed

    def _path_absolute(self, path):
        absolute = path
        if not os.path.isabs(path):
            if self.root:
                absolute = os.path.join(self.root, path)
            elif self.mode == "srv":
                absolute = os.path.join(self.parser.root, path)
            else:
                absolute = os.path.join(self.parser.clientconfdir, path)
        return absolute

    @property
    def _glob_changed(self):
        if not self._glob or not self._glob_dir:
            return False
        mtime = os.path.getmtime(self._glob_dir)
        return mtime != self._glob_mtime

    def extend(self):
        """Helper function for the parsing"""
        if not self._dirty and self.extended and not self._glob_changed:
            return self.extended
        paths = []
        root = self._path_absolute(self.value)
        for path in glob(root):
            if (
                self.parser._is_secure_path(path)
                and os.path.isfile(path)
                and not path.endswith("~")
                and not path.endswith(".back")
            ):
                paths.append(path)
        # more than one match mean we have a glob, for sure
        if len(paths) > 1 or (len(paths) == 1 and paths[0] != root):
            self._glob = True
            self._glob_dir = os.path.dirname(root)
            self._glob_mtime = os.path.getmtime(self._glob_dir)
        else:
            self._glob = False
        self.clean()
        self.extended = paths
        return paths

    def parse(self):
        """Parse the option value"""
        return self.extend()

    def dump(self):
        """Return the option representation to store in configuration file"""
        if self.extend() and not self.parser.backend.enforce:
            name = self.name or ""
            if self.is_template and name.startswith("../"):
                name = str(name[3:])
            return ". {}".format(name)
        # if the include did not match anything, we can safely remove it
        return ""


class OptionTpl(Option):
    """Option type Template

    Example:
        . .buitemplates/windows
    """

    type = "template"
    delim = ""

    def __init__(self, parser, name, value=None):
        """
        :param parser: Parser instance
        :type parser: :class:`burpui.misc.parser.burp1.Parser`
        """
        super(OptionTpl, self).__init__(name, value)
        self.parser = parser
        self.extended = False
        self._dirty = True
        if name:
            self._id = name.split(os.path.sep)[-1]
        else:
            self._id = ""

    @property
    def dirty(self):
        return self._dirty

    def _path_absolute(self, path):
        absolute = path
        if not os.path.isabs(path):
            absolute = os.path.join(self.parser.clientconfdir, path)
        return absolute

    def extend(self):
        """Helper function for the parsing"""
        if not self._dirty and self.extended:
            return self.extended
        path = self._path_absolute(self.value)
        self.clean()
        self.extended = path
        return path

    def parse(self):
        """Parse the option value"""
        return self.extend()

    def dump(self):
        """Return the option representation to store in configuration file"""
        if self.extend() and not self.parser.backend.enforce:
            return ". {}".format(self.name)
        # if the include did not match anything, we can safely remove it
        return ""


option_for_type = {
    "integer": OptionInt,
    "string": OptionStr,
    "boolean": OptionBool,
    "include": OptionInc,
}


class OptionMulti(Option):
    """Option type Multi

    Example:
        keep = 7
        keep = 4
    """

    type = "multi"

    def __init__(self, parser, name, value=None):
        self.parser = parser
        self.name = name
        self._dirty = True
        self._is_reset = []
        self.content_type = self.parser.advanced_type.get(name, "string")
        self.associate = getattr(self.parser, "pair_associations", {}).get(self.name)
        self._init_value(value)
        self.idx = len(self.value) - 1

    def _init_value(self, value):
        if value:
            container = self._obj_for_type()
            if not isinstance(value, list):
                value = container(self.name, value)
                self.value = [value]
            else:
                self.value = [container(self.name, x) for x in value]
        else:
            self.value = []

    def _obj_for_type(self):
        return option_for_type.get(self.content_type, OptionStr)

    def _wrap_object(self, value):
        container = self._obj_for_type()
        return container(self.name, value)

    def append(self, value, reset=None):
        self._dirty = True
        if isinstance(value, list):
            for v in value:
                self.value.append(self._wrap_object(v))
        else:
            self.value.append(self._wrap_object(value))
        if reset is not None:
            self.set_reset(reset)
        return self.value

    def update(self, value):
        """Change the option value"""
        self._dirty = True
        self._init_value(value)
        self._is_reset = []

    def remove(self, value):
        idx = self.value(value)
        del self.value[idx]
        self._dirty = True
        return self.value

    def index(self, value):
        for i, x in enumerate(self):
            if x == value:
                return i
        raise ValueError("{} is not in list".format(value))

    def len(self):
        return len(self.value)

    def dump(self, start=0, strict=True):
        """Return the option representation to store in configuration file"""
        ret = ""
        if start > len(self.value):
            return ret
        res = [self.dump_index(i, strict) for i in range(start, len(self.value))]
        ret = "\n".join(res)
        return ret.rstrip("\n")

    def dump_index(self, index, strict=True):
        if index >= len(self.value):
            return ""
        val = self.value[index]
        delim = self.delim
        self.idx = index
        if self.is_reset():
            delim = self.reset_delim
        return "{} {} {}".format(
            self.name, delim, sanitize_string("{}".format(val.parse()), strict)
        )

    def parse(self):
        return [x.parse() for x in self.value]

    def get_reset(self):
        return self._is_reset

    def __len__(self):
        return len(self.value)

    def __getitem__(self, ii):
        return self.value[ii].parse()

    def __delitem__(self, ii):
        del self.value[ii]
        self._dirty = True

    def __setitem__(self, ii, val):
        self.value[ii] = self._wrap_object(val)
        self._dirty = True

    def insert(self, ii, val):
        self.value.insert(ii, self._wrap_object(val))

    def __iter__(self):
        for x in self.value:
            yield x.parse()


class OptionPair(Option, dict):
    """Option type Pair

    Example:
        port = 4971
        max_children = 5
    """

    type = "pair"

    def __init__(self, parser, name, value=None):
        self.parser = parser
        self.name = name
        self.association = getattr(self.parser, "pair_associations", {}).get(self.name)
        self._dirty = True
        self._init_value(name, value)
        # is there a special := syntax for this option
        self._is_reset = []

    def _init_value(self, name, value):
        self.value = {}
        if value:
            value = OptionMulti(self.parser, name, value)
            self.value[name] = value

    def append(self, name, value):
        self._dirty = True
        if name != self.name:
            self.association = name
        if name in self.value:
            self.value[name].append(value)
        else:
            self.value[name] = OptionMulti(self.parser, name, value)
        return self.value

    def update(self, name, value):
        """Change the option value"""
        self._dirty = True
        self._init_value(name, value)
        self._is_reset = []

    def remove(self, name, value):
        self._dirty = True
        try:
            self.value.get(name, []).remove(value)
        except ValueError:
            pass
        return self.value

    def index(self, name, value):
        try:
            return self.value.get(name, []).index(value)
        except ValueError:
            return -1

    def len(self, name):
        if name in self.value:
            return self.value[name].len()
        return -1

    def dump(self, start=0, strict=True):
        """Return the option representation to store in configuration file"""
        ret = ""
        try:
            self_length = len(self.value[self.name])
        except KeyError:
            self_length = -1
        try:
            associate_length = len(self.value[self.association])
        except KeyError:
            associate_length = -1
        if start >= self_length and start >= associate_length:
            return ret
        for idx in range(start, max(self_length, associate_length)):
            v1 = self.dump_index(self.name, idx, strict)
            v2 = self.dump_index(self.association, idx, strict)
            if v1:
                ret += "{}\n".format(v1)
            if v2:
                ret += "{}\n".format(v2)
        return ret.rstrip("\n")

    def dump_index(self, name, index, strict):
        length = self.len(name)
        try:
            length = len(self.value[name])
        except KeyError:
            length = -1
        if index >= length:
            return ""
        return self.value.get(name).dump_index(index, strict)

    def parse(self, key=None):
        if key:
            if key not in self.value:
                return []
            return self.value[key].parse()
        ret = {}
        for key, opts in self.value.items():
            ret[key] = opts.parse()
        return ret

    def __setitem__(self, key, item):
        self.append(key, item)

    def __getitem__(self, key):
        return self.value.get(key, OptionMulti(self.parser, key))

    def get(self, key, default=None):
        try:
            return self.value[key]
        except KeyError:
            if default:
                return default
            return OptionMulti(self.parser, key)

    def __len__(self):
        return len(self.value)

    def __delitem__(self, key):
        del self.value[key]

    def clear(self):
        self.value.clear()

    def copy(self):
        return self.value.copy()

    def has_key(self, key):
        return key in self.value

    def keys(self):
        return self.value.keys()

    def values(self):
        return self.value.values()

    def items(self):
        return self.value.items()

    def pop(self, *args):
        return self.value.pop(*args)

    def __contains__(self, item):
        return item in self.value

    def __iter__(self):
        return iter(self.value)


class File(dict):
    """Object representing a configuration file

    :param parser: Parser object
    :type parser: :class:`burpui.misc.parser.doc.Doc`
    """

    md5 = None
    mtime = 0

    def __init__(self, parser, name=None, mode="srv", parent=None, is_template=False):
        """
        :param parser: Parser object
        :type parser: :class:`burpui.misc.parser.doc.Doc`

        :param name: File name
        :type name: str

        :param mode: Configuration type
        :type mode: str
        """
        # _dirty is used to know if the object changed
        self._dirty = False
        # _changed is used to know if the file changed since last read
        self._changed = True
        # _parsing_templates is used to know if we are currently parsing templates
        self._parsing_templates = False
        # cache the content of the file
        self._raw = []
        self._raw_data = MultiDict()
        self._data = MultiDict()
        self._is_template = is_template
        self.parser = parser
        self.mode = mode
        self.name = name
        self.parent = parent
        self.updated = []
        self.associations = set()
        self.reset = {}
        self.options = OrderedDict()
        self.types = {
            "boolean": OrderedDict(),
            "integer": OrderedDict(),
            "include": OrderedDict(),
            "multi": OrderedDict(),
            "pair": OrderedDict(),
            "string": OrderedDict(),
            "template": OrderedDict(),
        }
        if self.name:
            self.parse()

    @property
    def changed(self):
        for key, val in self.types["include"].items():
            if val.dirty:
                self._changed = True
                return self._changed
        for key, val in self.types["template"].items():
            if val.dirty:
                self._changed = True
                return self._changed
        try:
            if self.name:
                mtime = os.path.getmtime(self.name)
            else:
                self._changed = True
                return True
        except os.error:
            self._changed = True
            return True
        if mtime != self.mtime:
            self._changed = self._md5 != self.md5
            return self._changed
        self._changed = mtime != self.mtime
        return self._changed

    @property
    def version(self):
        return (
            getattr(self.parser.backend, "server_version", None)
            or getattr(self.parser.backend, "client_version", None)
            or ""
            if self.parser.backend
            else ""
        )

    def _ret_data(self, raw=True):
        if raw:
            ret = self._raw_data
        else:
            ret = self._data
        if ret and not self.dirty:
            ret.setlist("includes", [x for x in self.flatten("include", False).keys()])
            return ret
        ret.clear()
        for key, val in self.options.items():
            if isinstance(val, OptionMulti) and not raw:
                ret.setlist(key, val.parse())
            elif isinstance(val, OptionPair) and not raw:
                ret.setlist(key, val.parse(key))
            else:
                ret[key] = val if raw else val.parse()
        ret.setlist("includes", [x for x in self.flatten("include", False).keys()])
        ret.setlist("includes_ori", [x for x in self.flatten("include", False).keys()])
        return ret

    @property
    def raw_data(self):
        return self._ret_data()

    @property
    def data(self):
        return self._ret_data(False)

    def clone(self):
        cpy = File(self.parser, name=self.name, mode=self.mode)
        cpy.options = copy(self.options)
        cpy.types = copy(self.types)
        return cpy

    def clean(self):
        self._dirty = False
        for opt in self.options.values():
            opt.clean()

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name
        self.parse()

    def flatten(self, typ, listed=True, parse=True):
        self._refresh_types()
        if listed:
            return [
                {
                    "name": key,
                    "value": opt.parse() if parse else opt,
                    "reset": opt.get_reset(),
                }
                for key, opt in self.types[typ].items()
            ]
        ret = OrderedDict()
        if typ in self.types:
            for key, opt in self.types[typ].items():
                ret[key] = opt.parse() if parse else opt
        return ret

    def flatten_obj(self, name, obj, parse=True):
        return {
            "name": name,
            "value": obj.parse() if parse else obj,
            "reset": obj.get_reset(),
        }

    @property
    def dirty(self):
        if not self._dirty:
            self._dirty = any(x.dirty for x in self.options.values())
        return self._dirty

    @property
    def boolean(self):
        return self.flatten("boolean")

    @property
    def integer(self):
        return self.flatten("integer")

    @property
    def pair(self):
        return self.flatten("pair")

    @property
    def include(self):
        return self.flatten("include")

    @property
    def template(self):
        ret = []
        for tpl in self.flatten("template", parse=False):
            ret.append(
                {
                    "value": tpl["name"],
                    "name": tpl["value"]._id,
                }
            )
        return ret

    @property
    def multi(self):
        return self.flatten("multi")

    @property
    def string(self):
        return self.flatten("string")

    def _refresh_types(self):
        if self._dirty:
            for key in self.types.keys():
                self.types[key] = OrderedDict()

            for key, opt in self.options.items():
                if key not in self.associations:
                    self.types[opt.type][key] = opt

        self._dirty = False

    def _options_for_type(self, typ):
        return getattr(self.parser, "{}_{}".format(typ, self.mode), [])

    def _type_for_option(self, opt):
        if opt == ".":
            return "include"

        for typ in ["boolean", "integer", "multi", "string", "pair"]:
            if opt in self._options_for_type(typ):
                return typ
        return None

    def _new_opt(self, key, value=None, typ=None):
        typ = typ or self._type_for_option(key)

        if typ == "boolean":
            return OptionBool(key, value)
        if typ == "integer":
            return OptionInt(key, value)
        if typ == "multi":
            opt = self.options.get(key, OptionMulti(self.parser, key))
            opt.append(value)
            return opt
        if typ == "pair":
            return OptionPair(self.parser, key, value)
        if typ == "include":
            key = value
            return OptionInc(
                self.parser,
                key,
                value,
                root=self.name,
                mode=self.mode,
                template=self._is_template,
            )

        return OptionStr(key, value)

    @property
    def _md5(self):
        """Compute the md5sum of the file"""
        hash_md5 = md5()
        try:
            with open(self.name, "rb") as bfile:
                for chunk in iter(lambda: bfile.read(4096), b""):
                    hash_md5.update(chunk)
            return hash_md5.hexdigest()
        except IOError:
            return None

    def get(self, key, default=None):
        try:
            if key in self._options_for_type("pair"):
                return self.options[key].get(key)
            return self.options[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        if key in self._options_for_type("pair"):
            return self.options[key].get(key)
        return self.options[key]

    def __setitem__(self, key, value):
        self._dirty = True
        if key in self._options_for_type("boolean"):
            opt = OptionBool(key, value)
        elif key in self._options_for_type("integer"):
            opt = OptionInt(key, value)
        elif key in self._options_for_type("multi"):
            opt = self.options.get(key, OptionMulti(self.parser, key))
            if isinstance(value, list):
                opt.update(value)
            else:
                opt.append(value)
        elif key in self._options_for_type("pair"):
            association = self.parser.pair_associations.get(key)
            if key not in self.options and association not in self.options:
                self.associations.add(association)
                opt = OptionPair(self.parser, key)
            elif association in self.options:
                opt = self.options.get(association)
            else:
                opt = self.options.get(key)
            if isinstance(value, list):
                opt.update(key, value)
            else:
                opt.append(key, value)
        elif key == ".":
            key = value
            if self._parsing_templates:
                opt = OptionTpl(self.parser, key, value)
            else:
                opt = OptionInc(
                    self.parser,
                    key,
                    value,
                    root=self.name,
                    mode=self.mode,
                    template=self._is_template,
                )
                if self._is_template and key.startswith("../"):
                    key = str(key[3:])
        else:
            opt = OptionStr(key, value)
        self.options[key] = opt
        if key not in self.associations:
            self.types[opt.type][key] = opt

    def __repr__(self):
        self._refresh_types()
        ret = ""
        for key, opts in self.types.items():
            ret += "{} =>\n".format(key)
            for key2, opt in opts.items():
                if key2 in self.associations:
                    continue
                ret += "\t" + repr(opt) + "\n"
        return ret.rstrip("\n")

    def __str__(self):
        ret = ""
        for key, val in self.options.items():
            if key in self.associations:
                continue
            tmp = str(val)
            if tmp:
                ret += tmp + "\n"
        return ret.rstrip("\n")

    def __len__(self):
        return len(self.options)

    def __delitem__(self, key):
        self._dirty = True
        del self.options[key]

    def clear(self):
        self._dirty = True
        for key in self.types.keys():
            self.types[key].clear()
        return self.options.clear()

    def copy(self):
        return self.options.copy()

    def has_key(self, k):
        return k in self.options

    def update(self, *args, **kwargs):
        self._dirty = True
        return self.options.update(*args, **kwargs)

    def keys(self):
        return self.options.keys()

    def values(self):
        return self.options.values()

    def items(self):
        return self.options.items()

    def pop(self, *args):
        self._dirty = True
        return self.options.pop(*args)

    def __contains__(self, item):
        return item in self.options

    def __iter__(self):
        return iter(self.options)

    @property
    def raw(self):
        if self._raw and not self._changed and not self.changed:
            return self._raw
        if not self.name:
            return self._raw
        try:
            with codecs.open(self.name, "r", "utf-8", errors="ignore") as fil:
                self._raw = [x.rstrip("\n") for x in fil.readlines()]
        except IOError:
            return self._raw

        self._changed = False
        self.mtime = os.path.getmtime(self.name)
        self.md5 = self._md5
        return self._raw

    def _dump_resets(self):
        ret = {}
        for key, val in self.options.items():
            resets = val.get_resets()
            if resets:
                ret[key] = resets
        return ret

    def _format_key(self, key, data, index=None, dry=False):
        strict = "regex" not in key
        if not dry:
            self._changed = True

        # special case
        if key == ".":
            val = sanitize_string(data)
            if self._is_template and not val.startswith("../"):
                val = "../{}".format(val)
            if dry:
                return val
            else:
                return f". {val}"

        if index is not None and index >= 0:
            if key not in self.updated:
                val = data.getlist(key)
                if key in self:
                    self[key].update(val)
                else:
                    self[key] = val
                self.updated.append(key)
                if key in self.reset:
                    self[key].set_resets(self.reset[key])
            return str(self[key].dump_index(index, strict))

        if key in getattr(self.parser, "multi_{}".format(self.mode)) or key in getattr(
            self.parser, "pair_{}".format(self.mode), []
        ):
            if key not in self.updated:
                val = data.getlist(key)
                if key in self:
                    self[key].update(val)
                else:
                    self[key] = val
                self.updated.append(key)
                if key in self.reset:
                    self[key].set_resets(self.reset[key])
            val = self[key]
            return str(self[key].dump(strict=strict))

        if key not in self.updated:
            if key in self.parser.boolean_srv or key in self.parser.boolean_cli:
                val = data.get(key)
            else:
                val = sanitize_string(str(data.get(key)), strict)
            if key in self:
                self[key].update(val)
            else:
                self[key] = val
            self.updated.append(key)
            if key in self.reset:
                self[key].set_resets(self.reset[key])

        if dry:
            ret = self[key].parse()
            if index is not None:
                return ret[index]
            return ret
        return str(self[key])

    def parse(self, force=False):
        """Parse the current config"""
        if not self._changed and not self.changed and not force:
            return

        self.clear()
        for line in self.raw:
            if re.match(r"^\s*#", line):
                if BEGIN_TEMPLATES in line:
                    self._parsing_templates = True
                if END_TEMPLATES in line:
                    self._parsing_templates = False
                continue
            res = re.search(r"\s*([^=\s]+)\s*(:)?=?\s*(.*)$", line)
            if res:
                key = res.group(1)
                reset = res.group(2)
                val = res.group(3)
                if key == "compression":
                    val = val.replace("zlib", "gzip")
                elif key == "ssl_compression":
                    val = val.replace("gzip", "zlib")
                self[key] = val
                if key in self:
                    try:
                        self[key].set_reset(reset is not None)
                    except AttributeError:
                        pass

        self._dirty = False

    def _store(self, data, dest=None, insecure=False):
        """Store the config"""
        dest = dest or self.name
        if not dest:
            return [[NOTIF_ERROR, "No file defined!"]]

        dirname = os.path.dirname(dest)
        filename = os.path.basename(dest)
        if dirname and not os.path.exists(dirname):
            try:
                os.makedirs(dirname, 0o755)
            except OSError as exp:
                return [[NOTIF_WARN, str(exp)]]

        if not insecure:
            self.reset = {}
            ref = os.path.join(dirname, ".{}.bui.init.back~".format(filename))
            bak = os.path.join(dirname, ".{}.back~".format(filename))

            if not os.path.isfile(ref) and os.path.isfile(dest):
                try:
                    shutil.copy(dest, ref)
                except IOError as exp:
                    return [[NOTIF_ERROR, str(exp)]]
            elif os.path.isfile(dest):
                try:
                    shutil.copy(dest, bak)
                except IOError as exp:
                    return [[NOTIF_ERROR, str(exp)]]
        else:
            self.reset = self._dump_resets()

        def _make_it_bool(array):
            return ["{}".format(x).lower() == "true" for x in array]

        errs = []

        for key in data.keys():
            if key in self.parser.files:
                dat = data.get(key)
                if not os.path.isfile(dat):
                    typ = "strings"
                    if key in getattr(self.parser, "multi_{}".format(self.mode)):
                        typ = "multis"
                    elif key in getattr(self.parser, "boolean_{}".format(self.mode)):
                        typ = "bools"
                    elif key in getattr(self.parser, "integer_{}".format(self.mode)):
                        typ = "integers"
                    # highlight the wrong parameters
                    errs.append(
                        [
                            NOTIF_ERROR,
                            "Sorry, the file '{}' does not exist".format(dat),
                            key,
                            typ,
                        ]
                    )
            elif key.endswith(RESET_IDENTIFIER):
                target = key.replace(RESET_IDENTIFIER, "")
                if target in getattr(self.parser, "multi_{}".format(self.mode)):
                    self.reset[target] = _make_it_bool(data.getlist(key))
                else:
                    self.reset[target] = data.get(key)

        if errs and not insecure:
            return errs

        orig = self.raw
        oldkeys = [self._get_line_key(x) for x in orig]
        newkeys = list(set(data.keys()) - set(oldkeys))

        multi_index_map = {}
        pair_index_map = {}
        already_multi = set()
        already_pair = set()
        already_file = []
        written = []
        result = []
        self.updated = []

        def _lookup_option(key, val, start=0, strict=True, comment=True):
            """returns a list of tuples (idx, line)"""
            # re.match implies /^.../
            reg = r"\s*{}\s*:?=?".format(key)
            if comment:
                reg = r"\s*#*{}".format(reg)
            if strict:
                reg = r"{}\s*{}$".format(reg, val)
            start = min(start, len(orig))
            for idx, line in enumerate(orig[start:], start):
                if re.match(reg, line.rstrip("\n")):
                    return idx, line
            return -1, None

        def _is_key_after(key, start=0):
            """checks if a key is present in the following lines"""
            idx, _ = _lookup_option(key, None, start, False)
            return idx != -1

        def _dump(line, comment=None, raw=False):
            if raw:
                return line
            lead = ""
            if comment:
                lead = "#"
            return "{}{}".format(lead, line)

        try:
            with codecs.open(dest, "w", "utf-8", errors="ignore") as fil:
                # f.write('# Auto-generated configuration using Burp-UI\n')
                data_keys = list(data.keys())
                if "templates" in data:
                    result.append(_dump(" {}".format(BEGIN_TEMPLATES), True))
                    tpls = data.getlist("templates")
                    for tpl in tpls:
                        result.append(self._format_key(".", tpl))
                    result.append(_dump(" {}".format(END_TEMPLATES), True))
                skip_line = False
                for idx, line in enumerate(orig):
                    if self._line_is_comment(line) and BEGIN_TEMPLATES in line:
                        skip_line = True
                    if self._line_is_comment(line) and END_TEMPLATES in line:
                        skip_line = False
                        continue
                    if skip_line:
                        continue
                    key = self._get_line_key(line, False)
                    if (
                        self._line_removed(line, data_keys)
                        and not self._line_is_comment(line)
                        and not self._line_is_file_include(line)
                    ):
                        # The line was removed, we comment it
                        result.append(_dump(line, comment=True))
                    elif self._line_is_file_include(line):
                        # The line is a file inclusion, we check if the line
                        # was already present
                        ori = self._include_get_file(line)
                        if self._is_template and ori.startswith("../"):
                            ori = str(ori[3:])
                        if (
                            ori in data.getlist("includes_ori")
                            and ori not in already_file
                        ):
                            idx = data.getlist("includes_ori").index(ori)
                            inc = data.getlist("includes")[idx]
                            result.append(self._format_key(".", inc))
                            already_file.append(inc)
                        else:
                            if not insecure:
                                comment = not self._line_is_comment(line)
                                result.append(_dump(line, comment=comment))
                            else:
                                result.append(_dump(line))

                    elif key in data_keys:
                        # The line is still present or has been un-commented,
                        # rewrite it with eventual changes
                        multi = key in getattr(
                            self.parser, "multi_{}".format(self.mode)
                        )
                        pair = key in getattr(
                            self.parser, "pair_{}".format(self.mode), []
                        )
                        if pair and key not in pair_index_map:
                            pair_index_map[key] = 0
                        if multi and key not in multi_index_map:
                            multi_index_map[key] = 0
                        if key in written:
                            result.append(
                                _dump(line, comment=(not self._line_is_comment(line)))
                            )
                        else:
                            if multi:
                                length = len(self[key]) if key in self else -1
                                if key not in already_multi and (
                                    key not in self
                                    or (key in self and length > multi_index_map[key])
                                ):
                                    result.append(
                                        self._format_key(
                                            key, data, multi_index_map[key]
                                        )
                                    )
                                    multi_index_map[key] += 1
                                else:
                                    result.append(
                                        _dump(
                                            line,
                                            comment=(not self._line_is_comment(line)),
                                        )
                                    )
                                    continue
                                if len(self[key]) == multi_index_map[key]:
                                    already_multi.add(key)
                                    continue
                                # dump the rest of the multi if there are no
                                # more keys in the conf
                                if not _is_key_after(key, idx + 1):
                                    rest = self[key].dump(multi_index_map[key])
                                    if rest:
                                        result.append(_dump(rest))
                                    multi_index_map[key] = length
                                    already_multi.add(key)
                            elif pair:
                                length = len(self[key]) if key in self else -1
                                if key not in already_pair and (
                                    key not in self
                                    or (key in self and length > pair_index_map[key])
                                ):
                                    result.append(
                                        self._format_key(key, data, pair_index_map[key])
                                    )
                                    pair_index_map[key] += 1
                                else:
                                    result.append(
                                        _dump(
                                            line,
                                            comment=(not self._line_is_comment(line)),
                                        )
                                    )
                                    continue
                                if len(self[key]) == pair_index_map[key]:
                                    already_pair.add(key)
                                    continue
                                # dump the rest of the pair if there are no more
                                # keys in the conf
                                if not _is_key_after(key, idx + 1):
                                    rest = self[key].dump(pair_index_map[key])
                                    if rest:
                                        result.append(_dump(rest))
                                    pair_index_map[key] = length
                                    already_pair.add(key)
                            else:
                                # if key is not a multi and is not a password, we
                                # recycle the line
                                if key not in getattr(
                                    self.parser, "multi_{}".format(self.mode)
                                ) and (
                                    key == "password_check" or "password" not in key
                                ):
                                    if self._line_is_comment(line) and key in data:
                                        result.append(self._format_key(key, data))
                                        if key in newkeys:
                                            ridx = newkeys.index(key)
                                            del newkeys[ridx]
                                        continue
                                # The line was a comment and there was a further
                                # matching setting, so we just jump to the
                                # following
                                else:
                                    if self._line_is_comment(line) and _lookup_option(
                                        key, None, idx + 1, False
                                    ):
                                        result.append(_dump(line, raw=True))
                                        continue

                                val = self._format_key(key, data, dry=True)
                                lookup, _ = _lookup_option(key, val, idx + 1)
                                # The same option is here later, skip the
                                # current one
                                if lookup != -1:
                                    result.append(_dump(line, raw=True))
                                    continue

                                written.append(key)
                                result.append(self._format_key(key, data))
                    else:
                        result.append(
                            _dump(
                                line,
                                comment=(
                                    key in written and not self._line_is_comment(line)
                                ),
                            )
                        )

                # write the rest of the multi settings
                for key, idx in multi_index_map.items():
                    if key not in already_multi and idx < self[key].len():
                        result.append(_dump(self[key].dump(idx)))
                # write the rest of the pair settings
                for key, idx in pair_index_map.items():
                    if key not in already_pair and idx < self[key].len():
                        result.append(_dump(self[key].dump(idx)))
                # Write the rest of file inclusions
                if "includes" in data:
                    for inc in data.getlist("includes"):
                        if inc not in already_file:
                            result.append(self._format_key(".", inc))
                # Write the new keys
                for key in newkeys:
                    if key.endswith(RESET_IDENTIFIER):
                        continue
                    if (
                        key not in written
                        and key not in already_multi
                        and key not in already_pair
                        and key not in ["includes", "includes_ori", "templates"]
                    ):
                        result.append(self._format_key(key, data))

                if len(result) > 0 and result[-1] != "":
                    result.append("")
                fil.write("\n".join(result))

        except Exception as exp:
            # something went wrong, we revert to the last file
            with codecs.open(dest, "w", "utf-8", errors="ignore") as fil:
                fil.writelines(orig)
            return [[NOTIF_ERROR, str(exp)]]

        self.parse(True)

        return [[NOTIF_OK, "Configuration successfully saved."]]

    def store(self, dest=None, insecure=False):
        """Store the current conf object"""
        data = self.data
        return self._store(data, dest, insecure)

    def store_data(self, data, insecure=False):
        """Store the conf given a object data"""
        return self._store(data, insecure=insecure)

    @staticmethod
    def _line_is_comment(line):
        """Check whether a given line is a comment or not"""
        if not line:
            return False
        return line.startswith("#")

    @staticmethod
    def _line_is_file_include(line):
        """Check whether a given line is a file inclusion or not"""
        if not line:
            return False
        return line.startswith(".") or re.match(r"^#+\s*\.", line) is not None

    @staticmethod
    def _include_get_file(line):
        """Return the path of the included file(s)"""
        if not line:
            return None
        _, fil = re.split(r"\s+", line, 1)
        return fil

    @staticmethod
    def _get_line_key(line, ignore_comments=True):
        """Return the key of a given line"""
        if not line:
            return ""
        if "=" not in line:
            return line
        (key, _) = re.split(r"\s*:?=\s*", line, 1)
        if not ignore_comments:
            key = key.strip("# ")
        return key.strip()

    @staticmethod
    def _line_removed(line, keys):
        """Check whether a given line has been removed in the updated version"""
        if not line:
            return False
        (key, _) = re.split(r"\s+|:?=", line, 1)
        key = key.strip()
        return key not in keys


class Config(File):
    """Object representing a configuration

    A config is like a virtual file so we can reuse some methods

    :param parser: Parser object
    :type parser: :class:`burpui.misc.parser.doc.Doc`
    """

    def __init__(self, path=None, parser=None, mode="srv"):
        """
        :param parser: Parser object
        :type parser: :class:`burpui.misc.parser.doc.Doc`

        :param mode: Configuration type
        :type mode: str
        """
        # we need an OrderedDict since the order of the configuration matters
        self.files = OrderedDict()
        self._tree = []
        self.default = path
        self.name = path
        self._includes = []
        self._templates = []
        self._is_template = False
        self._dirty = True
        if path:
            self.files[path] = File(parser, path, mode=mode)
        super(Config, self).__init__(parser, path, mode)

    @property
    def changed(self):
        for path, conf in self.files.items():
            if conf.changed:
                return True
        return False

    def _parse(self):

        orig = self.files.copy()
        for root, conf in orig.items():
            conf.parse()
            for key, val in conf.flatten("include", False).items():
                for path in val:
                    if not os.path.isabs(path):
                        path = os.path.join(os.path.dirname(root), path)
                    self.add_file(path, root)
                    self._includes.append(path)
            for key, path in conf.flatten("template", False).items():
                if not os.path.isabs(path):
                    path = os.path.join(os.path.dirname(root), path)
                self.add_file(path, root)
                self._templates.append(path)

        # recursively parse the conf
        if orig != self.files:
            self._parse()

    def parse(self, force=False):
        if not self.changed and not force:
            return

        del self._includes[:]
        del self._templates[:]
        self._parse()

        removed = []
        orig = self.files.copy()
        for path, conf in orig.items():
            if conf.parent and (
                (conf.name not in self._includes and conf.name not in self._templates)
                or conf.name in removed
            ):
                removed.append(path)
                self.del_file(path)

    @property
    def tree(self):
        if not self.changed and not self.dirty and self._tree:
            return self._tree

        # make sure to refresh files list
        self.parse(True)

        def __new_node(name, parent=None):
            dirname = os.path.dirname(name)
            basename = os.path.basename(name)
            return {
                "name": basename,
                "title": basename,
                "full": name,
                "dir": dirname,
                "parent": parent,
                "children": [],
            }

        self._tree[:]
        dflt = self.get_default(True)
        temp = {}

        # retrieve the offset of the default conf
        offset = 0
        for idx, path in enumerate(self.files.keys()):
            if path == dflt.name:
                offset = idx
                break

        for idx, (top, conf) in enumerate(self.files.items()):
            if idx < offset:
                continue
            if idx > offset:
                break
            node = __new_node(conf.name)
            for key, val in conf.flatten("include", False).items():
                for path in val:
                    if not os.path.isabs(path):
                        path = os.path.join(os.path.dirname(top), path)
                    node["children"].append(__new_node(path, node["full"]))
            temp[conf.name] = node
        self._tree = [x for x in temp.values()]
        return self._tree

    def store(self, conf=None, dest=None, insecure=False):
        ret = []
        if conf and conf in self.files:
            return self.files[conf].store(dest, insecure)
        for name, conf in self.files.items():
            ret += conf.store(insecure=insecure)
        return ret

    def store_data(self, conf, data, insecure=False):
        return self.get_file(conf).store_data(data, insecure)

    def clone(self):
        cpy = Config(self.name, self.parser, self.mode)
        for path, parsed in self.files.items():
            if path == self.name:
                continue
            cpy.add_file(path)
        return cpy

    def set_default(self, path):
        self.default = path
        self.name = path
        try:
            default = self.get_default(True)
            if not self.parser and default:
                self.parser = getattr(self.get_default(), "parser")
            if not self.mode and default:
                self.mode = getattr(self.get_default(), "mode")
        except ValueError:
            pass

    def get_default(self, exc=False):
        if self.default:
            return self.get_file(self.default)
        if exc:
            raise ValueError("No default configuration found")
        return File(self.parser, mode=self.mode)

    def add_file(self, path=None, parent=None):
        idx = path or self.default
        if idx not in self.files:
            self.files[idx] = File(
                self.parser, idx, self.mode, parent, self._is_template
            )
            self._dirty = True
        return self.files[idx]

    def get_file(self, path):
        ret = self.files.get(path, File(self.parser, path, mode=self.mode))
        ret.parse()
        return ret

    def del_file(self, path):
        self._dirty = True
        del self.files[path]

    def list_files(self):
        return self.files.keys()

    def set_template(self, val):
        self._is_template = val

    def _refresh(self):
        if self._dirty or any(x.dirty for x in self.files.values()):
            # cleanup "caches"
            self.options.clear()
            self.options = OrderedDict()
            for key in list(self.types.keys()):
                del self.types[key]
                self.types[key] = OrderedDict()

            # now update caches with new values
            for fil in self.files.values():
                self.options.update(fil.options)
                self.associations = self.associations.union(fil.associations)
                # FIXME: find a way to cache efficiently
                # fil.clean()

            for key, val in self.options.items():
                if key not in self.associations:
                    self.types[val.type][key] = val

        self._dirty = False

    def _get(self, key, default=None, raw=False):
        self._refresh()
        try:
            obj = self.options[key]
            if obj.type in ["pair"] and not raw:
                obj = obj.get(key)
        except KeyError:
            if default:
                return default
            if self.parser and key in self.parser.defaults:
                obj = self._new_opt(key, self.parser.defaults[key])
            else:
                return None
        return obj if raw else obj.parse()

    def get_raw(self, key, default=None):
        return self._get(key, default, True)

    def get(self, key, default=None):
        return self._get(key, default)

    def getlist(self, key):
        obj = self._get(key, raw=True)
        if not obj:
            return []
        if isinstance(obj, OptionMulti):
            return obj.parse()
        else:
            return [obj.parse()]

    def __getitem__(self, key):
        self._refresh()
        if key in self._options_for_type("pair"):
            return self.options[key].get(key)
        return self.options[key]

    def __setitem__(self, key, value):
        self.get_default(True)[key] = value
        self._dirty = True

    def __repr__(self):
        self._refresh()
        ret = ""
        for key, fil in self.files.items():
            ret += ">" * 5 + key + "<" * 5 + "\n"
            ret += repr(fil) + "\n"
        return ret.rstrip("\n")

    def __str__(self):
        self._refresh()
        return super(Config, self).__str__()

    def __len__(self):
        self._refresh()
        return len(self.options)

    def __delitem__(self, key):
        del self.get_default(True)[key]
        self._dirty = True

    def clear(self):
        self._dirty = True
        return self.files.clear()

    def copy(self):
        return self.files.copy()

    def has_key(self, k):
        self._refresh()
        return k in self.options

    def update(self, *args, **kwargs):
        self._dirty = True
        return self.get_default(True).update(*args, **kwargs)

    def keys(self):
        self._refresh()
        return self.options.keys()

    def values(self):
        self._refresh()
        return self.options.values()

    def items(self):
        self._refresh()
        return self.options.items()

    def pop(self, *args):
        self._dirty = True
        return self.get_default(True).pop(*args)

    def __contains__(self, item):
        self._refresh()
        return item in self.options

    def __iter__(self):
        self._refresh()
        return iter(self.options)



# -*- coding: utf8 -*-
"""
.. module:: burpui.engines.agent
    :platform: Unix
    :synopsis: Burp-UI agent module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import struct
import sys
import ssl
import json
import logging
import time

import trio

from functools import partial

from ..exceptions import BUIserverException
from ..misc.backend.interface import BUIbackend
from .._compat import pickle, to_bytes, to_unicode
from ..config import config
from ..desc import __version__

# TODO: sendfile is not yet supported by trio
# try:
#     from sendfile import sendfile
#     USE_SENDFILE = True
# except ImportError:
#     USE_SENDFILE = False


BUI_DEFAULTS = {
    "Global": {
        "port": 10000,
        "bind": "::",
        "ssl": False,
        "sslcert": "",
        "sslkey": "",
        "backend": "burp2",
        "password": "azerty",
        "init_wait": 15,
    },
}


class BurpHandler(BUIbackend):
    # These functions MUST be implemented because we inherit an abstract class.
    # The hack here is to get the list of the functions and let the interpreter
    # think we don't have to implement them.
    # Thanks to this list, we know what function are implemented by our backend.
    foreign = BUIbackend.__abstractmethods__
    BUIbackend.__abstractmethods__ = frozenset()

    def __init__(self, backend="burp2", logger=None, conf=None, init_wait=0):
        self.backend_name = backend
        self.is_async = backend == "parallel"
        self.logger = logger
        wait = init_wait

        top = __name__
        if "." in self.backend_name:
            module = self.backend_name
        else:
            if "." in top:
                top = top.split(".")[0]
            module = "{0}.misc.backend.{1}".format(top, self.backend_name)
        try:
            sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
            mod = __import__(module, fromlist=["Burp"])
            if self.is_async:
                Client = mod.AsyncBurp
            else:
                Client = mod.Burp
            self.backend = Client(conf=conf)

            def __backend_alive():
                stats = self.backend.statistics()
                return "alive" in stats and stats["alive"]

            alive = __backend_alive()
            while not alive and wait > 0:
                self.logger.debug(
                    "Waiting for the backend to become alive... {}/{}".format(
                        init_wait - wait, init_wait
                    )
                )
                time.sleep(1)
                alive = __backend_alive()
                wait -= 1
            if not alive:
                raise BUIserverException("Cannot talk to burp server")
        except Exception as exc:
            self.logger.error(
                "Failed loading backend {}: {}".format(self.backend_name, str(exc)),
                exc_info=exc,
                stack_info=True,
            )
            sys.exit(2)

    def __getattribute__(self, name):
        # always return this value because we need it and if we don't do that
        # we'll end up with an infinite loop
        if name in ["foreign", "backend", "logger", "is_async"]:
            return object.__getattribute__(self, name)
        # now we can retrieve the 'foreign' list and know if the object called
        # is in the backend
        if name in self.foreign:
            return getattr(self.backend, name)
        try:
            return getattr(self.backend, name)
        except AttributeError:
            pass
        return object.__getattribute__(self, name)


class BUIAgent(BUIbackend):
    BUIbackend.__abstractmethods__ = frozenset()

    def __init__(self, conf=None, level=0, logfile=None):
        self.padding = 1
        level = level or 0
        self.logger.init_logger(config=dict(level=level, logfile=logfile))
        lvl = self.logger.getEffectiveLevel()
        self.logger.info("conf: {}".format(conf))
        self.logger.info("level: {}".format(logging.getLevelName(lvl)))
        if not conf:
            raise IOError("No configuration file found")

        # Raise exception if errors are encountered during parsing
        self.conf = config
        self.conf.parse(conf, BUI_DEFAULTS)
        self.conf.default_section("Global")
        self.port = self.conf.safe_get("port", "integer")
        self.bind = self.conf.safe_get("bind")
        self.backend = self.conf.safe_get("backend")
        self.ssl = self.conf.safe_get("ssl", "boolean")
        self.sslcert = self.conf.safe_get("sslcert")
        self.sslkey = self.conf.safe_get("sslkey")
        self.password = self.conf.safe_get("password")
        self.init_wait = self.conf.safe_get("init_wait", "integer")
        self.conf.setdefault("BUI_AGENT", True)

        self.client = BurpHandler(self.backend, self.logger, self.conf, self.init_wait)

    def _ssl_context(self):
        if not self.ssl:
            return None
        ctx = ssl.SSLContext()
        ctx.load_cert_chain(self.sslcert, self.sslkey)
        return ctx

    async def run(self):
        try:
            self.logger.debug(f"Starting server on {self.bind}:{self.port}")
            ctx = self._ssl_context()
            if ctx:
                await trio.serve_ssl_over_tcp(
                    self.handle, self.port, ctx, host=self.bind
                )
            else:
                await trio.serve_tcp(self.handle, self.port, host=self.bind)
        except KeyboardInterrupt:
            self.logger.debug("Stopping server")
            sys.exit(0)

    async def handle(self, server_stream: trio.StapledStream):
        """self.request is the client connection"""
        self.logger.debug("handle request")
        try:
            err = None
            res = ""
            lengthbuf = await server_stream.receive_some(8)
            if not lengthbuf:
                return
            (length,) = struct.unpack("!Q", lengthbuf)
            data = await self.receive_all(server_stream, length)
            self.logger.info(f"recv: {data!r}")
            txt = to_unicode(data)
            if txt == "RE":
                return
            j = json.loads(txt)
            if j["password"] != self.password:
                self.logger.warning("-----> Wrong Password <-----")
                await server_stream.send_all(b"KO")
                return
            try:
                if j["func"] == "proxy_parser":
                    parser = self.client.get_parser()
                    if j["args"]:
                        wrap = partial(getattr(parser, j["method"]), **j["args"])
                    else:
                        wrap = getattr(parser, j["method"])
                    temp = await trio.to_thread.run_sync(wrap)
                    res = json.dumps(temp)
                elif j["func"] == "agent_version":
                    res = json.dumps(__version__)
                elif j["func"] == "restore_files":
                    wrap = partial(getattr(self.client, j["func"]), **j["args"])
                    if self.client.is_async:
                        res, err = await wrap()
                    else:
                        res, err = await trio.to_thread.run_sync(wrap)
                    if err:
                        await server_stream.send_all(b"ER")
                        await server_stream.send_all(struct.pack("!Q", len(err)))
                        await server_stream.send_all(to_bytes(err))
                        self.logger.error("Restoration failed")
                        return
                elif j["func"] == "get_file":
                    path = j["path"]
                    path = os.path.normpath(path)
                    err = None
                    if not path.startswith("/"):
                        err = f"The path must be absolute! ({path})"
                    if not path.startswith(self.client.tmpdir):
                        err = f"You are not allowed to access this path: " f"({path})"
                    if err:
                        await server_stream.send_all(b"ER")
                        await server_stream.send_all(struct.pack("!Q", len(err)))
                        await server_stream.send_all(to_bytes(err))
                        self.logger.error(err)
                        return
                    count = 0
                    size = os.path.getsize(path)
                    await server_stream.send_all(b"OK")
                    await server_stream.send_all(struct.pack("!Q", size))
                    async with await trio.open_file(path, "rb") as f:
                        while True:
                            buf = await f.read(1024)
                            if not buf:
                                break
                            buflen = len(buf)
                            count += buflen
                            percent = count / size * 100
                            self.logger.info(f"sending {buflen} Bytes - {percent:.1f}%")
                            await server_stream.send_all(buf)
                    os.unlink(path)
                    lengthbuf = await server_stream.receive_some(8)
                    (length,) = struct.unpack("!Q", lengthbuf)
                    data = await self.receive_all(server_stream, length)
                    txt = to_unicode(data)
                    if txt == "RE":
                        return
                elif j["func"] == "del_file":
                    path = j["path"]
                    path = os.path.normpath(path)
                    err = None
                    if not path.startswith("/"):
                        err = f"The path must be absolute! ({path})"
                    if not path.startswith(self.client.tmpdir):
                        err = f"You are not allowed to access this path: " f"({path})"
                    if err:
                        berr = to_bytes(err)
                        await server_stream.send_all(b"ER")
                        await server_stream.send_all(struct.pack("!Q", len(berr)))
                        await server_stream.send_all(berr)
                        self.logger.error(err)
                        return
                    res = json.dumps(False)
                    if os.path.isfile(path):
                        os.unlink(path)
                        res = json.dumps(True)
                else:
                    callback = getattr(self.client, j["func"])
                    if j["args"]:
                        if "pickled" in j and j["pickled"]:
                            # de-serialize arguments if needed
                            import hmac
                            import hashlib
                            from base64 import b64decode

                            pickles = to_bytes(j["args"])
                            key = "{}{}".format(self.password, j["func"])
                            key = to_bytes(key)
                            bytes_pickles = pickles
                            digest = hmac.new(
                                key, bytes_pickles, hashlib.sha1
                            ).hexdigest()
                            if not hmac.compare_digest(digest, j["digest"]):
                                raise BUIserverException(
                                    "Integrity check failed: {} != {}".format(
                                        digest, j["digest"]
                                    )
                                )
                            # We need to replace the burpui datastructure
                            # module by our own since it's the same but
                            # burpui may not be installed
                            mod = __name__
                            if "." in mod:
                                mod = mod.split(".")[0]
                            data = b64decode(pickles)
                            data = data.replace(
                                b"burpui.datastructures",
                                to_bytes(f"{mod}.datastructures"),
                            )
                            j["args"] = pickle.loads(data)

                        wrap = partial(callback, **j["args"])
                        if self.client.is_async:
                            res = json.dumps(await wrap())
                        else:
                            res = json.dumps(await trio.to_thread.run_sync(wrap))
                    else:
                        if self.client.is_async:
                            res = json.dumps(await callback())
                        else:
                            res = json.dumps(await trio.to_thread.run_sync(callback))
                self.logger.info(f"result: {res}")
                await server_stream.send_all(b"OK")
            except (BUIserverException, Exception) as exc:
                await server_stream.send_all(b"ER")
                res = str(exc)
                self.logger.error(res, exc_info=exc)
                self.logger.warning(f"Forwarding Exception: {res}")

            res = to_bytes(res)
            await server_stream.send_all(struct.pack("!Q", len(res)))
            await server_stream.send_all(res)
        except AttributeError as exc:
            self.logger.warning(f"Wrong method => {exc}", exc_info=exc)
            await server_stream.send_all(b"KO")
        except Exception as exc:
            self.logger.error(f"!!! {exc} !!!", exc_info=exc)

    async def receive_all(self, stream: trio.StapledStream, length=1024, bsize=None):
        buf = b""
        bsize = bsize if bsize is not None else 1024
        bsize = min(bsize, length)
        received = 0
        tries = 0
        while received < length:
            newbuf = await stream.receive_some(bsize)
            if not newbuf:
                # 3 successive read failure => raise exception
                if tries > 3:
                    raise Exception("Unable to read full response")
                tries += 1
                await trio.sleep(0.1)
                continue
            # reset counter
            tries = 0
            buf += newbuf
            received += len(newbuf)
        return buf



"""increase pref value size

Revision ID: 695dcbd29d4f
Revises: 56de018f4d88
Create Date: 2018-10-02 13:56:57.660204

"""

# revision identifiers, used by Alembic.
revision = "695dcbd29d4f"
down_revision = "56de018f4d88"

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pref", schema=None) as batch_op:
        batch_op.alter_column(
            "value",
            existing_type=sa.VARCHAR(length=256),
            type_=sa.Text(),
            existing_nullable=True,
        )

    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pref", schema=None) as batch_op:
        batch_op.alter_column(
            "value",
            existing_type=sa.Text(),
            type_=sa.VARCHAR(length=256),
            existing_nullable=True,
        )

    ### end Alembic commands ###
